(* extended-double-float *)

let echo = Array.length Sys.argv > 1;;

let extended_image n = (
	let data = Buffer.create 0 in
	Dfm.Writer.write_raw_extended (Buffer.add_substring data) n;
	let s = Buffer.contents data in
	let image = String.create (2 * String.length s) in
	for i = 0 to String.length s - 1 do
		let b = int_of_char s.[i] in
		image.[String.length image - i * 2 - 1] <- (Printf.sprintf "%x" (b land 0x0f)).[0];
		image.[String.length image - i * 2 - 2] <- (Printf.sprintf "%x" (b lsr 4)).[0]
	done;
	image
);;

let test_ei s v = (
	let i = extended_image v in
	if echo then Printf.printf "%s\n" i;
	i = s
);;

if echo then Printf.printf "**** write extended ****\n";;

assert (test_ei "00000000000000000000" 0.0);;
assert (test_ei "3fff8000000000000000" 1.0);;
assert (test_ei "40008000000000000000" 2.0);;
assert (test_ei "4000c000000000000000" (3.0));;
assert (test_ei "3ffe8000000000000000" (0.5));;
assert (test_ei "bfff8000000000000000" (-1.0));;
assert (test_ei "80000000000000000000" (-0.0));;

let extended_value image = (
	assert (String.length image = 20);
	let s = String.create 10 in
	for i = 0 to 9 do
		let b = int_of_string ("0x" ^ String.sub image (i * 2) 2) in
		s.[9 - i] <- char_of_int b
	done;
	let read data cursor dest index length = (
		let really_length = (
			if !cursor + length <= String.length data then length else
			String.length data - !cursor
		) in
		String.blit data !cursor dest index really_length;
		cursor := !cursor + really_length;
		really_length
	) in
	Dfm.Reader.read_raw_extended (read s (ref 0))
);;

let test_ev s be = (
	let v, _ = extended_value s in
	let vimg = Int64.bits_of_float v in
	let bimg = Int64.bits_of_float be in
	if echo then Printf.printf "(%.16Lx) %.16Lx %f\n" bimg vimg v;
	vimg = bimg
);;

if echo then Printf.printf "**** read extended ****\n";;

assert (test_ev "00000000000000000000" 0.0);;
assert (test_ev "3fff8000000000000000" 1.0);;
assert (test_ev "40008000000000000000" 2.0);;
assert (test_ev "4000c000000000000000" (3.0));;
assert (test_ev "3ffe8000000000000000" (0.5));;
assert (test_ev "bfff8000000000000000" (-1.0));;
assert (test_ev "80000000000000000000" (-0.0));;

(* date *)

let test_time datetime unixtime = (
	if echo then Printf.printf "datetime = %f\n" datetime;
	if echo then Printf.printf "unixtime = %f\n" unixtime;
	let u_of_d = Dfm.unixtime_of_datetime datetime in
	let d_of_u = Dfm.datetime_of_unixtime unixtime in
	if echo then Printf.printf "datetime of unixtime = %f\n" d_of_u;
	if echo then Printf.printf "unixtime of datetime = %f\n" u_of_d;
	assert (u_of_d = unixtime);
	assert (d_of_u = datetime)
);;

test_time 36692.0 961027200.0;; (* 2000-06-15 00:00:00.000 *)
test_time 36692.56250 961075800.0;; (* 2000-06-15 13:30:00.000 *)

(* serialize *)

type variant = A | B;;

let string_of_variant (x: variant): string = (
	match x with
	| A -> "A"
	| B -> "B"
);;

let variant_of_string (x: string): variant = (
	match x with
	| "A" -> A
	| "B" -> B
	| _ -> failwith "variant_of_string"
);;

let io_variant ?key v s = Serialize.make_io
	(fun x -> `tag (string_of_variant x))
	(fun x -> match x with `tag s -> variant_of_string s | _ -> failwith "io_variant")
	?key v s;;

type t = {w: variant; x: float; y: string; z: char option};;

let io_t ?key v s = Serialize.io_record (fun v s -> {
	w = io_variant ~key:"w" v.w s;
	x = Serialize.io_float ~key:"x" v.x s;
	y = Serialize.io_string ~key:"y" v.y s;
	z = Serialize.io_option Serialize.io_char '\x00' ~key:"z" v.z s}) ?key v s;;

let data1 = {w = A; x = 1.0; y = "hello"; z = Some 'A'};;
let data2 = {w = B; x = -0.5; y = ""; z = None};;

let pp_float2 ?(pp = Serialize_pp.pp_pervasives) f v = (
	match v with
	| `float v ->
		Format.pp_print_float f v;
		if v = floor v then Format.pp_print_char f '0'
	| `custom_test ->
		assert false
	| a ->
		pp f a
);;

(* pretty printer *)
let w = Serialize_pp.create ~pp:pp_float2 Format.str_formatter in
let (_: t) = io_t data1 w in
let (_: int list) = Serialize.io_list Serialize.io_int 0 [1; 2; 3; 4; 5] w in
let s = Format.flush_str_formatter () in
if echo then Printf.printf "%s" s;
assert (s = "{w = A; x = 1.0; y = \"hello\"; z = Some \'A\'}[1; 2; 3; 4; 5]\n");;

let w = Serialize_pp.create Format.str_formatter in
let (_: t) = io_t data2 w in
let (_: int list) = Serialize.io_list Serialize.io_int 0 [] w in
let s = Format.flush_str_formatter () in
if echo then Printf.printf "%s" s;
assert (s = "{w = B; x = -0.5; y = \"\"; z = None}[]\n");;

(* dfm *)
let f = open_out_bin "./test.dfm" in
let w = Dfm.Writer.create "TEST" f in
let (_: t) = io_t data1 w in
close_out f;;

let f = open_in_bin "./test.dfm" in
let r = Dfm.Reader.create "TEST" f in
let v = io_t data2 r in
let (_: unit) = assert (v = data1) in
close_in f;;

let f = open_out_bin "./test.dfm" in
let w = Dfm.Writer.create "TEST" f in
let (_: t) = io_t data2 w in
close_out f;;

let f = open_in_bin "./test.dfm" in
let r = Dfm.Reader.create "TEST" f in
let v = io_t data1 r in (* None should not be recorded. *)
let (_: unit) = assert (v = {data2 with z = data1.z}) in
close_in f;;

(* complex structure *)

let stream = [
	"", `record;
		"R1R", `record;
			"A", (`int 1);
			"B", (`int 2);
			"RRR", `record;
				"C", (`int 3);
			"", `end_of_record;
		"", `end_of_record;
		"R2L", `list `any;
		"", `end_of_list;
		"R3L", `list `any;
			"", (`int 4);
			"", (`int 5);
		"", `end_of_list;
		"R4L", `list `any;
			"", `list `any;
				"", (`int 6);
				"", (`int 7);
			"", `end_of_list;
			"", `list `any;
				"", (`int 8);
				"", (`int 9);
			"", `end_of_list;
		"", `end_of_list;
		"R5L", `list `any;
			"", `record;
				"D", (`int 10);
				"RLRR", `record;
					"E", (`int 11);
				"", `end_of_record;
			"", `end_of_record;
		"", `end_of_list;
		"F", (`int 12);
	"$", `end_of_record];;

let f = open_out_bin "./test.dfm" in
let w = Dfm.Writer.create "TEST" f in
List.iter (fun (key, value) ->
	Serialize.write w ~key:key value
) stream;
close_out f;;

let f = open_in_bin "./test.dfm" in
let r = Dfm.Reader.create "TEST" f in
List.iter (fun (key, value) ->
	let k, v = Serialize.peek r in
	if key <> "$" then Serialize.junk r;
	if echo then (
		let pp = Serialize_pp.pp_serialize ~pp:Serialize_pp.pp_pervasives in
		Format.printf "(%s) %a : (%s) %a@." k pp v key pp value
	);
	assert (k = key || key = "$");
	assert (v = value);
) stream;
close_in f;;
